##################################################################
#
# XLRteams
# team-shuffling and balancing plugin for b3
# (c) 2011 Evan King (Honoredsoft@gmail.com)
#
# This program is free software and licensed under the terms of
# the GNU General Public License (GPL), version 2.
#
##################################################################

from __future__ import with_statement
from threading import Lock, RLock
__author__  = 'Evan King'
__version__ = '0.9.0'

# TODO: ignore "canrebalancethisplayer" checks when player count is under X
# TODO: implement option to move players not immediately, but the next time they die

# history:
# 0.9.0: added player team assignment on connect (beta)
# 0.8.1: added quick hack to skip maps when server empty
# 0.8.0: overhauled player score calculations
# 0.7.2: switched from threaded execution to sheduler (hoping to avoid untraceable but persistent thread safety issue)
# 0.7.1: team balances moving more than 50% of active players now invert switch selection to reduce actual moves
#   (...doesn't actually work, until the code is refactored so the in-progress lists are contained in an object that can be manipulated)
# 0.7.0: heavily reworked team state management--reliability checks are less spammy and pay less attention to unreliable b3 data
#   client disconnects/spectating bots/etc. now more clearly and consistently handled
# 0.6.4: try to fix b3's belief that many bots are spectating (unreliable, b3 reverts to unknown or spectating)
# 0.6.3: fixed broken event handling for players leaving
#   improved reliability of player team maintenance with more spectator exceptions
# 0.6.2: added exception handling for attempts to register EVT_GAME_MAP_CHANGE on pre 1.7.0 b3 systems
#   reverted the sync-breaking changes from 0.5.6, found a possible cause for the constant "x != x, moving x->x" team applications
# 0.6.1: during balance searches, players are now always allowed to be "moved" when the move actually puts them back on the team they currently occupy
# 0.6.0: added new config values: rebalance_disable_mapstart_seconds and rebalance_disable_mapend_seconds
#   added config option: override_pa_teams (when players issue !teams, convert to !balance)
# 0.5.6: code review tweaks--may have improved reliability, may have reintroduced player list drift
# 0.5.5: added cleanup on spectating bots (if they exist for any reason, xlrteams kicks them--fix for misbehaving bot support)
# 0.5.4: added B3's new EVT_GAME_MAP_CHANGE event handling
# 0.5.3: fixed deadlock in periodic balance checking thread (caused by erratic handling of XlrPlayer.checkTeam() calls)
# 0.5.2: added checks which prevent losing team sync
# 0.5.1: fixed how kick and drop events are handled (the team switch was treated as manual, which probably put them back outside "TEAM_DROPPED")
# 0.5.0: getPlayer now enforces constant sanity checks against out-of-sync player cids.
# 0.4.12: removed all connection triggers and limited disconnect triggers to players on teams (now trigger on team joins only)
#   started rebuilding player lists on restart and new map start (to fix sync issues with team reblance actions interrupted by map load)
# 0.4.11: switched balance requirement from 3+ per team to 6+ total (now that counts don't have to be so close), and added "not enough players" message
# 0.4.10: added kick/ban/tempban event handling
#   added team kill event handling (translates skill-wise to death for both)
# 0.4.9: more attempts added to overcome B3's bot confusion (players sharing cids with departed bots)
#   replaced !listbots with !listplayers, which lists red bots, blue bots, red players, and blue players separately
# 0.4.8: replaced static on-creation "are you a bot" check with dynamic derivation on the base client data
#   (workaround for B3 temporarily assigning null/broadcast address to clients)
#   added/improved reporting detail a bit
#   added !listbots command to verify humans are not being identified as bots
# 0.4.7: added player/client count overview to !balancereport
# 0.4.6: tightened auto-balance skill ratio
#   added validation that the new balance is an improvement (skill difference is smaller ratio)
#   improved balance result messaging (shorter fail messages, clarify how many switched players were bots)
#   loosened player count balancing portion of algorithm (that disruption caused most of the "too many iterations" failures)
# 0.4.5: bots no longer count toward maximum moves per balance operation
#   manual/end-of-map balance applications are much less likely to fail
#   the number of /human/ players allowed to be switched in a balance is reduced to about 35% of total players (humans+bots)
# 0.4.4: more check-prevention bugfixes, (restored equal weight xlrstats values, average now only happens if xlrstats values exist)
# 0.4.3: tweaked skill calculation (exaggerated skrstats-calculated skill component, removed skill boundaries for intermediate calculations)
# 0.4.2: fixed regression causing balance checks to start at beginning of a map
# 0.4.1: bots entering/leaving the game no longer trigger a balance check.  They will still be forced onto a team individually.
#   players spectating do trigger a balance check, but delayed 60 seconds, allowing time for a short reprieve or for checking a suspicious player
# 0.4.0: balance adjustments will no longer be allowed within the last couple minutes of a map (2 * balanceCooldownSeconds)
#   added a "stable player" factor to frequent switch prevention (avoid switching players who stick around a long time on one team)
#       - count balancing, which occurs before skill balancing, now also checks against the teamswitch prevention factor
#       - bounce prevention is now only applied when moving a player /away/ from his current team (not when the algorithm is putting him back)
#   prevented balance checking when a player spectates, so high-skill admins can get away with watching other players occasionally and not trigger a shakeup
#       - note this will only work if said spectating doesn't trigger the spawning of a new bot which will trigger a balance check anyway
#   skill calculation no longer uses xlrstats skill, but rather a calculation based on xlrstats recorded kills and deaths (the same as used for map skill calculation)
# 0.3.9: changed frequent switch prevention to always allow switching bots (slightly improves avoidance of switching humans)
#   fixed bug allowing rebalances affecting more than 55% of the players on the first try (should be only allowed after 2 consecutive fails by this reason)
#   more algorithmic tweaks and bugfixes which marginally improve the chance of finding balance and reduce the average number of switches per rebalance (or at least accepting an adjustment)
#   unbalanceable situations should now be handled more gracefully, first reporting impasse twice, eventually ensuring some change is applied for the sake of progress
#   loosened the allowed player count differences (both numeric and percentile) to better accomodate skill balancing of heterogenous rosters
# 0.3.8: introduced delayed handling of disconnects, so multiple disconnects at once trigger a single, effective rebalancing.
# 0.3.7: disallowed findBalance from failing on "too many moves" more than 3 consecutive times
# 0.3.6: tweaked skill calculating and autochecking factors (increasing value of star players)
# 0.3.5: fixed potential issue with the new move processing thread and non-team games
#   restored checkActive() call which ensures players exist (was accidentally removed)
#   improved granularity of activity checks (now only performed in team games, damage events update activity)
# 0.3.4: players are now forced to stay on an assigned team for at least two minutes (will switch them back if they try to change)
#      - this also fixes an issue where lamers would switch back within 15 seconds and leave grossly imbalanced teams for the maximum check interval
#   players on unknown teams are now assigned to whichever is smaller right away
#       - this also fixes an issue where a constant flow of new bots would make every rebalance force-applied and move many players
#   team changes are now only announced across center screen if human players are being moved
# 0.3.3: fixed circumstance where rebalancing could assign teams with counts differing by > 20% with small player counts
#   reverted changes from 0.3.1 due to bad behavior
# 0.3.2: rebalancing is no longer announced if only bots are being moved
# 0.3.1: restored some leniency in balancing team counts while balancing skill (after first balancing count to within +/- 1 player)
# 0.3.0: overhauled thread control, and forked actual team switches into another thread with a one-item-per-player queue
#   added more reporting info and some minor cosmetic tweaks
#   rebalancing now works on b3 restart without a map switch
# 0.2.2: improved/fixed some balance check-related timings, tweaked algorithm some more
# 0.2.1: added !balancereport command to provide some assessment feedback (for further algorithm testing)
#   further tweaked balance algorithm
# 0.2.0: tweaked balance algorithm
# 0.1.6: activity assertions ignore bots and occur less often, and bots have -500 skill
# 0.1.5: players which leave a game are now kept in memory until the map ends (prevents gaming balance by score reset)
#   fixed bug: the presence of spectators was causing balance attempts to behave as if there were players in unknown teams
#       - the result was requiring that balance be assigned even if it wasn't optimal
#   fixed bug: the setting "balance_prevent_bouncing" was not being correctly read from the xml config
#   failed adjustment attempts now prevent a repeat attempt for half the min-interval
# 0.1.4:
#   fixed bug: setTeam failing on manual switch (essentially broke team switching altogether)
#   fixed bug: manual checks weren't always being allowed within half the min-interval, even though they should have
#   improved player shuffling: disabled player switch prevention when performing a rebalance as part of a shuffle
#   improved timing of allowing manual checks (now halved /only/ for players with authlevel 60+)
#   improved timing of first allowed check (now max(60s, min-interval) rather than min-interval plus 60s)
#   improved timing of first automatic check  now max(90s, min-interval) rather than min-interval)
#   improved rebalancing: made rebalancing refuse to move much more than 50% of the total player count (rebalancing does nothing and fails)
#       - may not reliably work, and will not prevent rebalancing when shuffling or when there are players in unknown teams
# 0.1.3: made applyTeams move one player per side at a time to prevent UrT refusing some switches
# 0.1.2: added probabalistically preventing frequent switches of the same player, and configuration to toggle this behavior
# 0.1.1: added dummy player checks (players with no kills or deaths for the last 5 minutes are forced to spectate)
#    Helps deal with server crashes of which B3's plugins are unaware, and possibly player timeouts as well
# 0.1.0: initial release

import random, math, time, thread, threading
import b3
import b3.events

# STATE_UNKNOWN = 3, TEAM_BLUE = 3, TEAM_FREE = 0, TEAM_RED = 2, TEAM_SPEC = 1, TEAM_UNKNOWN = -1
#b3.TEAM_NONE = 4    # TEAM_UNKNOWN that is actually distinct from TEAM_SPEC
b3.TEAM_DROPPED = 5 # disconnected users (retained in memory until map ends)

# SOME API NOTES FOR FUTURE USE:
# self.console.clients.getByCID(k)
# self.console.clients.getByGUID(k)
# self.console.storage.query(sql)
# self.console.clients.getList()
# self.console.getPlayerList()
# self.console.game.gameType // returns short strings: dm tdm ts ftl cah ctf bm
# deathmatch, team dm, team survivor, follow the leader, capture and hold, capture the flag, bomb
# self.console.getPlugin('xlrstats')

# self.console.info(msg)
# self.console.debug(msg)
# self.console.warning(msg)
# self.console.error(msg)
# self.console.verbose(msg)

#Thread that triggers a task every N seconds (no longer used, now switched to scheduling due to a rare untraceable thread safety issue)
class PeriodicExecutor(threading.Thread):

    def __init__(self,func,sleep,params=[]):
        self.func = func
        self.params = params
        self.sleep = sleep
        threading.Thread.__init__(self,name = "PeriodicExecutor")
        self.setDaemon(1)
    
    def run(self):
        while 1:
            try:
                apply(self.func,self.params)
            except:
                pass
            finally:
                time.sleep(self.sleep)

class XlrPlayer(object):
    
    # config variables
    logScale = 2 # larger number reduces positive skill factor (numbers >10 will produce negligible skill factors)
    baseSkill = 1500 # base "average" skill (should match XLRStats base skill+typical top skill over 2)
    minSkill = 400
    maxSkill = 6000
    playerstatsTable = 'xlr_playerstats'
    playerstatsOffset = -200 #scale all players stats? (for tweaking relative balance-- i.e. (2-1):(3-1) is bigger difference than 2:3)
    xlrSkillDifferenceFactor = 3 #directly magnify the stat differences relative to the average typical score (baseSkill)
    commandTimeout = 5 # keep trying to send rcon commands until they're 5 seconds old
    guid = None
    
    _cid = -1
    _client = None
    _name = "Null*"
    _human = False
    _kills = 0
    _deaths = 0
    _skill = 1000
    _xlrskill = None
    _xlrid = 0
    _console = None
    _teamkeys = ["free", "s", "red", "blue", "free", "s"]
    _teamnames = ["^3Free^7", "^7Spectate^7", "^1Red^7", "^4Blue^7", "^5None^7", "^6Dropped^7"]
    # STATE_UNKNOWN = 3, TEAM_BLUE = 3, TEAM_FREE = 0, TEAM_RED = 2, TEAM_SPEC = 1, TEAM_UNKNOWN = 1
    _team = b3.TEAM_SPEC
    _bteam = None
    _teamlocked = False
    _queuedswitch = None
    _control = RLock()
    _lastaction = time.time()
    _lastcheck = time.time()
    # recent switching factor for auto switching prevention
    _lastbalanceswitch = 0
    _lastswitch = 0
    # loyalty factor for auto switching prevention
    _lastside = b3.TEAM_SPEC
    _lastsidejoin = 0
    _specbotcount = 0
    
    def __init__(self, console, client=None, startTeam=b3.TEAM_SPEC):
        self._console = console
        if client:
            self._client = client
            self.guid = client.guid
            self._cid = int(float(client.cid))
            self._bteam = startTeam
            #if client.team != b3.TEAM_UNKNOWN:
            self._team = client.team
            self._queuedswitch = None
            self._name = client.name
            self.message = self.doMessage
            self._xlrid = client.id
            self._lastaction = time.time()
            self._lastbalanceswitch = 0
            self._xlrskill = self.baseSkill
            self.reset()
            console.debug("Creating new client: %s[%s] @%s (%s) (%s) with xklrskill=%s" % (client.name, self.getTeamName(), client._ip, client.guid, client._team, self._xlrskill))
        else: # don't have local access to verbose reporting, have to find global method reference
            console.verbose("Object is not a client: %s" % repr(client))
            self.message = self.doNothing
    
    def __cmp__(self, other):
        if isinstance(other, XlrPlayer):
            skill = other.getSkill()
        else:
            skill = 0
        return skill - self.getSkill()
    
    def doMessage(self, msg):
        try:
            self.getClientObj().message(msg)
        except:
            pass
    
    def isBot(self):
        try:
            if self._human:
                return False
            client = self.getClientObj()
            # try: #client.bot only recently defined, and clearly not actually set
                # if client and not client.bot:
                    # self._human = True
                    # return False
            # except:
                # pass
            if client._ip not in ["0.0.0.0", "::0.0.0.0", "::", ""]:
                self._human = True
                return False
            return True
        except:
            return False
    
    def setHuman(self):
        self._human = True
        return self
    
    def doNothing(self):
        return self
    
    def checkActive(self):
        if self._xlrskill == self.baseSkill:
            # xlrteams failed to load the real skill--don't know why, but need to brute-force correct (repeat until the stupid is gone)
            self.loadXlrSkill()
        with self._control:
            if self.getTeam() in [b3.TEAM_SPEC, b3.TEAM_DROPPED]:
                #if self.isBot():
                #    self._console.write("clientkick %s You appear to be a spectating bot." % self._cid, self.commandTimeout)
                return self
            # player hasn't done anything measurable in 3 minutes, and probably doesn't exist or is at least inactive
            if time.time() - self._lastaction > 180: # and not self.isBot():
                # bots tend to disappear without us knowing, but we mustn't actually move them to spec
                if self.isBot():
                    self._queuedswitch = None
                    self._team = b3.TEAM_DROPPED
                else:
                    self._queuedswitch = b3.TEAM_DROPPED
            #else:
            #    self._console.say("checkActive: %s is active" % self._name)
            return self
    
    def getClientObj(self):
        if self._client:
            return self._client
        self._client = self._console.clients.getByGUID(self.guid)
        if not self._client:
            self._client = self._console.clients.getByExactName(self._name)
        if not self._client:
            self._client = self._console.clients.getByCID(self._cid)
        if not self._client:
            if self._team != b3.TEAM_DROPPED:
                self._console.error("Couldn't find client record for %s (%s, %s)" % (self._name, self._cid, self.guid))
                self._queuedswitch = b3.TEAM_DROPPED
                #self._console.write("clientkick %s client not found" % self._cid, self.commandTimeout)
            raise Exception("No client match")
            
        self.guid = self._client.guid
        self._cid = int(float(self._client.cid))
        self._name = self._client.name
        self._xlrid = self._client.id
        return self._client
    
    def getId(self):
        try:
            return self.getClientObj().cid
        except:
            return self._cid
    
    def getClientTeam(self):
        try:
            return self.getClientObj()._team
        except:
            return b3.TEAM_UNKNOWN
    
    def setClientTeam(self, tid=b3.TEAM_UNKNOWN):
        try:
            client = self.getClientObj()
            client._team = tid
            return True
        except:
            return False
    
    # attempt to detect and correct failures by UrT to respect/remember team assignments
    def checkTeam(self):
    
        # stopgap against not being able to actually (accurately) check state against the game server, but needing to avoid excessive team assignment spamming
        now = time.time()
        if now - self._lastcheck < 30 or self._queuedswitch:
            return self
        self._lastcheck = now
        
        with self._control:
            if self._queuedswitch:
                return self
            tid = int(float(self._team)) # int casting can fail on float string
            ctid = int(float(self.getClientTeam()))
            if tid == b3.TEAM_SPEC and self.isBot():
                self._queuedswitch = b3.TEAM_DROPPED
            elif tid in [b3.TEAM_RED, b3.TEAM_BLUE, b3.TEAM_SPEC] and ctid != tid:
                self._console.write("forceteam %s %s" % (self.getId(), self._teamkeys[tid]), self.commandTimeout)
                if ctid in [b3.TEAM_RED, b3.TEAM_BLUE]:
                    self.message("^7You may have missed a switch to the %s team (moving now, %s->%s)" % (self._teamnames[tid], ctid, tid))
                self._console.verbose("^7Moving %s to the %s team (moving now, %s->%s)" % (self._name, self._teamnames[tid], ctid, tid))
                self.setClientTeam(tid)
            # if self._team == b3.TEAM_DROPPED and self._client:
                # self.message("^5B3 is kicking you because it thinks you are a spectating bot.")
                # self._console.debug("^5B3 is kicking %s as spectating bot (%s|%s, %s)." % (self._name, tid, ctid, self.guid))
                # self._console.write("clientkick %s dropped" % self.getId(), self.commandTimeout)
                # self._client = None
                    
            return self
    
    def setTeam(self, tid, manual=True):
        with self._control:
            curtime = time.time()
            if tid == b3.TEAM_DROPPED:
                self._team = tid
                self._queuedswitch = None
                self._client = None
                return self
            if manual: # client moved himself or was moved by another plugin
                if tid in [b3.TEAM_RED, b3.TEAM_BLUE] and not self.isBot():
                    if self._bteam != tid and curtime - self._lastbalanceswitch < 120:
                        self.message("^5Don't be a spoil-sport.  Stay on your assigned team.")
                        self._queuedswitch = self._bteam
                self._team = tid
                self.act(True)
                return self
            if tid == self._queuedswitch:
                return False
            if self._team == tid:
                self.checkTeam()
                return False
            # keep track of how long a player has been loyal to this team
            if tid in [b3.TEAM_RED, b3.TEAM_BLUE] and self._lastside != tid:
                self._lastsidejoin = curtime
                self._lastside = tid
            self._bteam = tid
            self._console.debug("Balance setting %s to %s" % (self._name, self._teamkeys[tid]))
            # don't rebalance someone destined for inactivity drop
            if self._queuedswitch == b3.TEAM_DROPPED and not manual:
                return False
            if self._queuedswitch == tid:
                return False
            self._queuedswitch = tid
        return self
    
    def applyBalance(self):
        with self._control:
            if not self._queuedswitch:
                return False
            curtime = time.time()
            tid = self._queuedswitch
            #delay = abs(curtime - self._lastswitch)
            #if delay < 6:
            #    time.sleep(6 - delay)
            
            #if not self.isBot():
            if tid == b3.TEAM_DROPPED:
                if self.isBot():
                    self.message("^5B3 is kicking you because it thinks you are a spectating bot.")
                    self._console.verbose("%s (%s) now %s->dropped due to inactivity" % (self._name, self.guid, self._teamkeys[self._team]))
                    self._console.write("clientkick %s specbot" % self._cid, self.commandTimeout)
                    self._client = None
                else:
                    self._console.verbose("%s (%s) now %s->spectating due to inactivity" % (self._name, self.guid, self._teamkeys[self._team]))
                    self.message("^5You are now spectating due to inactivity.")
                self._client = False
                tid = b3.TEAM_SPEC
            else:
                if self._team == b3.TEAM_UNKNOWN:
                    self.message("^5B3 doesn't know what team you are on.")
                    tid = self._bteam
            if tid == b3.TEAM_UNKNOWN:
                return False
            self.message("^7You are being moved to the %s team" % self._teamnames[tid])
            self._lastswitch = curtime
            self._team = tid
            self._console.write("forceteam %s %s" % (self.getId(), self._teamkeys[tid]), self.commandTimeout)
            self.setClientTeam(tid)
            self._lastcheck = curtime
            self._lastsidejoin = curtime
            self._lastside = tid
            self._console.verbose("%s (%s) moved to team %s" % (self._name, self.getId(), self.getTeamName()))
            self._queuedswitch = None
            self._lastbalanceswitch = curtime
        return True
    
    def kill(self):
        self.act()
        self._kills += 1
        self._calcSkill()
        return self
    
    def die(self):
        self._deaths += 1
        self._calcSkill()
        return self
    
    # fromSpec means the client action didn't necessarily have to be performed from a team
    def act(self, fromSpec=False):
        self._lastaction = time.time()
        if fromSpec:
            return self
        if not self.getTeam() in [b3.TEAM_FREE, b3.TEAM_RED, b3.TEAM_BLUE]:
            if self._bteam:
                with self._control:
                    self._queuedswitch = self._bteam
            else:
                self._team = b3.TEAM_UNKNOWN # on next check, client will get auto-assigned a new team
        return self
        
    def reset(self):
        self._kills = 0
        self._deaths = 0
        self.loadXlrSkill()
        self._calcSkill()
        return self
    
    def loadXlrSkill(self):
        if self._xlrid == 0:
            if self._client and self._client.id != 0:
                self._xlrid = self._client.id
            else:
                return self
        q = 'SELECT kills, deaths, skill from %s WHERE client_id = %s LIMIT 1' % (self.playerstatsTable, self._xlrid)
        cursor = self._console.storage.query(q)
        if cursor and (cursor.rowcount > 0):
            r = cursor.getRow()
            self._xlrskill = r['skill'] + self.playerstatsOffset
            #self._xlrskill = self.kdrScore(r['kills'], r['deaths']) + self.playerstatsOffset
        return self
    
    def getTeam(self):
        with self._control:
            if self._queuedswitch:
                return self._queuedswitch
            if self._team != b3.TEAM_UNKNOWN:
                return self._team
            self._team = self.getClientTeam()
            return self._team

    def getTeamName(self, team=None):
        if team == None:
            team = self.getTeam()
        
        if team == b3.TEAM_FREE:
            return "free"
        if team == b3.TEAM_RED:
            return "red"
        if team == b3.TEAM_BLUE:
            return "blue"
        if team == b3.TEAM_UNKNOWN:
            return "unknown"
        if team == b3.TEAM_DROPPED:
            return "dropped"
        if team == b3.TEAM_SPEC:
            return "spectating"
        return "other"
    
    def canRebalance(self):
        # bots can always be kicked around as often as we please
        if self.isBot():
            return True
        # putting someone back on their current team actually reduces moves
        if (self._team == b3.TEAM_RED and self._queuedswitch == b3.TEAM_BLUE) or (self._team == b3.TEAM_BLUE and self._queuedswitch == b3.TEAM_RED):
            return True
        if self._teamlocked:
            return False
        elapsed = time.time() - self._lastbalanceswitch
        duration = max(time.time() - self._lastsidejoin, 765) # (900-765) / 900 = 0.15, 15% chance
        if elapsed < 120: # never switch within 2 minutes of last switch
            return False
        # at 7m 30s, probability of allowing switch reaches 100%
        prob = elapsed / 450
        # at 15m, it reaches 0% (but capped at 15%)
        prob = min(prob, 0.2 + (900 - duration)/900)
        #probability is much higher than perceived, because checks will occur on multiple passes per rebalance
        return True if random.random() <= prob else False
    
    def allowRebalance(self):
        self._lastbalanceswitch = 0
    
    def getLastAdjust(self):
        return self._lastbalanceswitch
    
    def getName(self):
        return self._name
    
    def setName(self, name):
        self._console.verbose("%s (%s) renamed to %s" % (name, self.getId(), self._name))
        self._name = name
        self.act(True)
        return self
    
    def getSkill(self):
        return self._skill
    
    def _calcSkill(self):
        if self._xlrskill == self.baseSkill:
            # xlrteams failed to load the real skill--don't know why, but need to brute-force correct (repeat until the stupid is gone)
            self.loadXlrSkill()
            
        skill = self.kdrScore(self._kills, self._deaths)
        # average map skill against (doubled difference) xlrstats skill
        if self._xlrskill != self.baseSkill:
            # magnify the xlrstats skill difference, because we need knapstack-style relationships
            xlrskill = self.baseSkill + self.xlrSkillDifferenceFactor * (self._xlrskill - self.baseSkill)
            #xlrskill = self._xlrskill - (self.baseSkill - self._xlrskill)
            skill = (skill + 2*xlrskill) / 3
        skill = min(max(skill, self.minSkill), self.maxSkill)
        if self.isBot():
            skill = skill - 400
        self._skill = skill
        return skill
    
    def kdrScore(self, kills, deaths):
        k = kills+2 if kills > 0 else 2 # kdr applies newcomers a score below xlrskill factor unless scaled a little higher--diminishing curve makes this fine at higher kdr's
        d = float(deaths) if deaths > 0 else 1.0
        if k > d:
            ratio = math.log(self.logScale - 1 + (k / d), self.logScale)
        else:
            ratio = (k / d) * (0.75)
        return ratio * self.baseSkill
        #return min(max(ratio*self.baseSkill, self.minSkill), self.maxSkill)

# encapsulation of a balance attempt
class XlrBalance(object):
    redlist = {}
    bluelist = {}
    speclist = {}
    redskill = 0
    blueskill = 0
    state = "normal" # eor (end of round), manual, etc.
    
    def __init__(self, playerlist, state=None):
        # build lists according to current teams (getTeam() returns queued changes over current values)
        for k, p in playerlist.items():
            if p.getTeam() == b3.TEAM_BLUE:
                self.bluelist.append(p)
                self.blueskill += p.getSkill()
            elif p.getTeam() in [b3.TEAM_RED, b3.TEAM_UNKNOWN]: # capturing unknowns here
                self.redlist.append(p)
                self.redskill += p.getSkill()
        if state:
            self.state = state

class XlrteamsPlugin(b3.plugin.Plugin):
    debugMode = False
    debugBalance = False

    autoShuffle = True
    autoBalance = True
    overridePATeams = False
    balancePreventBouncing = True
    balanceCooldownSeconds = 60
    balanceIntervalSeconds = 300
    balanceIntervalRounds = 5
    balanceDisableSecondsBOM = 60
    balanceDisableSecondsEOM = 90
    minPlayers = 4
    
    # "auto" values trigger automatic rebalancing if enabled, but the balancer always tries to get within the manual balance check parameters
    maxTeamDiff = 3
    maxAutoTeamDiff = 4
    
    maxTeamRatio = .25
    maxAutoTeamRatio = .35
    
    maxSkillRatio = .1
    maxAutoSkillRatio = .2
    immediateSkillRatio = .75
    
    maxIdleMap = 2 * 60
    
    _players = {} # player id => XlrPlayer()
    _nullplayer = None
    _adjustresult = "none"
    _checkresult = "none"
    _botshuffling = False
    _lastlistupdate = 0
    _lastadjust = time.time()
    _lastcheck = time.time()
    _nextcheck = time.time() + 600
    _toomanycount = 0
    _toolongcount = 0
    _control = RLock()
    _checkthread = None
    _switchthread = None
    _emptysince = 0
    
    _mapstart = time.time()
    _roundstart = time.time()
    _lastreset = time.time()
    
    # load XML configuration
    def onLoadConfig(self):
        try:
            self.autoShuffle = self.config.getboolean('balance', 'auto_shuffle')
        except:
            self.debug('Using default value (%s) for auto_shuffle', self.autoShuffle)
        try:
            self.autoBalance = self.config.getboolean('balance', 'auto_balance')
        except:
            self.debug('Using default value (%s) for auto_balance', self.autoBalance)
            
        try:
            self.balanceCooldownSeconds = self.config.getint('balance', 'rebalance_interval_min_seconds')
        except:
            self.debug('Using default value (%s) for rebalance_interval_min_seconds', self.balanceCooldownSeconds)
            
        try:
            self.balanceIntervalSeconds = self.config.getint('balance', 'rebalance_interval_max_seconds')
        except:
            self.debug('Using default value (%s) for rebalance_interval_max_seconds', self.balanceIntervalSeconds)
        
        try:
            self.balanceIntervalRounds = self.config.getint('balance', 'rebalance_interval_rounds')
        except:
            self.debug('Using default value (%s) for rebalance_interval_rounds', self.balanceIntervalRounds)
        
        try:
            self.balancePreventBouncing = self.config.getboolean('balance', 'prevent_player_bouncing')
        except:
            self.debug('Using default value (%s) for prevent_player_bouncing', self.balancePreventBouncing)
        
        try:
            self.balanceDisableSecondsBOM = self.config.getint('balance', 'rebalance_disable_mapstart_seconds')
        except:
            self.debug('Using default value (%s) for rebalance_disable_mapstart_seconds', self.balanceDisableSecondsBOM)
        
        try:
            self.balanceDisableSecondsEOM = self.config.getint('balance', 'rebalance_disable_mapend_seconds')
        except:
            self.debug('Using default value (%s) for rebalance_disable_mapend_seconds', self.balanceDisableSecondsEOM)
        
        try:
            self.overridePATeams = self.config.getboolean('balance', 'override_pa_teams')
        except:
            self.debug('Using default value (%s) for override_pa_teams', self.overridePATeams)
        
        self._nextcheck = time.time() + max(90, self.balanceCooldownSeconds)
        self._checkthread = PeriodicExecutor(self.intervalBalanceCheck, 10).start()
        # self.console.cron + b3.cron.PluginCronTab(self, self.intervalBalanceCheck, '*/10')
        self._switchthread = PeriodicExecutor(self.processMoveQueue, 3).start()
        # self.console.cron + b3.cron.PluginCronTab(self, self.processMoveQueue, '*/3')

    # startup calls after onLoadConfig
    def startup(self):
        # sink object for method calls on null player (rather than endless try/catch blocks)
        self._nullplayer = XlrPlayer(self.console)
        
        # add current players
        for c in self.console.clients.getList():
            if c.cid >= 0:
                self.getPlayer(c)
            else:
                self.debug("found player %s[%s] (%s)" % (c.name, c.team, c.cid))

        # get the admin plugin so we can register commands
        admin = self.console.getPlugin('admin')
        # register our commands
        commands = ["shuffle", "autoshuffle", "autobalance", "preventbouncing", "balancereport", "listplayers", "scores"]
        try:
            level = self.config.getint('balance', 'authlevel')
            self.debug('Using configured value (%s) for authlevel', level)
        except:
            level = 60
            self.debug('Using default value (%s) for authlevel', level)
        # admin.registerCommand(self, cmd, level, func, alias)
        for c in commands:
            admin.registerCommand(self, "xlr"+c, level, self.getCmd(c), c)
        admin.registerCommand(self, "xlrbalance", 20, self.getCmd("balance"), "balance")
        admin.registerCommand(self, "xlrteams", 20, self.getCmd("balance"))
        if self.overridePATeams:
            admin.registerCommand(self, "pateams", 20, self.getCmd("balance"), "teams")
        
        # register the events we're interested in. 
        self.registerEvent(b3.events.EVT_CLIENT_CONNECT) # command is spammy with garbage, so we'll rely on on-the-fly creations from other events
        self.registerEvent(b3.events.EVT_CLIENT_DISCONNECT)
        self.registerEvent(b3.events.EVT_CLIENT_KICK)
        self.registerEvent(b3.events.EVT_CLIENT_BAN)
        self.registerEvent(b3.events.EVT_CLIENT_BAN_TEMP)
        self.registerEvent(b3.events.EVT_CLIENT_JOIN)
        self.registerEvent(b3.events.EVT_CLIENT_TEAM_CHANGE)
        self.registerEvent(b3.events.EVT_CLIENT_NAME_CHANGE)
        self.registerEvent(b3.events.EVT_CLIENT_KILL_TEAM)
        self.registerEvent(b3.events.EVT_CLIENT_KILL)
        self.registerEvent(b3.events.EVT_CLIENT_SUICIDE)
        self.registerEvent(b3.events.EVT_GAME_ROUND_START)
        self.registerEvent(b3.events.EVT_GAME_ROUND_END)
        self.registerEvent(b3.events.EVT_GAME_WARMUP)
        self.registerEvent(b3.events.EVT_GAME_EXIT)
        try: # throws an exception on pre-1.7.0 b3 systems which don't have this event
            self.registerEvent(b3.events.EVT_GAME_MAP_CHANGE)
        except:
            pass
        #self.registerEvent(b3.events.EVT_CLIENT_DAMAGE)
        #self.registerEvent(b3.events.EVT_CLIENT_DAMAGE_SELF)
        
        self.updatePlayerList(True)
    
    def onEvent(self, event):
        evts = b3.events
        type = event.type
        curtime = time.time()
        try:
            p = self.getPlayer(event.client)
        except:
            p = self.getPlayer()
        
        if type == evts.EVT_CLIENT_CONNECT:
            # add player to list UNLESS we already have him (actually already done now)
            #self.getPlayer(event.client)
            # save the check for the join
            #if not p.isBot():
            #    self._nextcheck = min(curtime + 30, self._nextcheck)
        #elif type in [evts.EVT_CLIENT_DISCONNECT, evts.EVT_CLIENT_KICK, evts.EVT_CLIENT_BAN, evts.EVT_CLIENT_BAN_TEMP]:
        elif type == evts.EVT_CLIENT_DISCONNECT:
            #try:
            
            #c = self._players[event.data]
            c = self.findPlayer(event.data)
            self.verbose("disconnected: %s (%s)" % (c.getName(), event.data))
            c.setTeam(b3.TEAM_DROPPED, False)
            # schedule a sheck soon instead of executing one right away, so multiple disconnects close together will be more effectively handled (ignore bots)
            if not c.isBot() and c.getTeam() in [b3.TEAM_BLUE, b3.TEAM_RED]:
                self._nextcheck = min(curtime + 10, self._nextcheck)
            
            # except:
                # self.verbose("disconnected: (unknown) (%s)" % event.data)
                # for k, p in self._players.items():
                    # self.verbose("    GUID: %s, CID: %s, NAME %s" % (p.guid, p._cid, p._name))
                # pass
        elif type == evts.EVT_CLIENT_JOIN or type == evts.EVT_CLIENT_TEAM_CHANGE:
            # move client to correct team
            p.setTeam(event.client.team)
            # if last adjustment time is within 15 seconds, this was probably a team change event generated /by/ that adjustment
            # so, we don't want to run checkBalance again, because even though it wouldn't do an adjustment, it would still shorten
            # the time to next balance check from the normal maximum interval
            # other reasons not to check:
            #   - if the person joined the spectators, he might be coming back shortly
            #   - if it's a bot, it could be an automatic reaction to another change we should ignore, and probably doesn't make a noteworthy difference
            if curtime - self._lastadjust > 15 and p and not p.isBot():
                delay = 5 if event.client.team in [b3.TEAM_RED, b3.TEAM_BLUE] else 60
                self._nextcheck = min(curtime + delay, self._nextcheck)
        elif type == evts.EVT_CLIENT_NAME_CHANGE:
            p.setName(event.client.name)
        elif type in [evts.EVT_CLIENT_DAMAGE, evts.EVT_CLIENT_DAMAGE_SELF]:
            p.act();
        elif type == evts.EVT_CLIENT_KILL:
            if event.client == event.target:
                return
            p.kill()
            self.getPlayer(event.target).die()
        elif type == evts.EVT_CLIENT_KILL_TEAM:
            p.die().act()
            self.getPlayer(event.target).die()
        elif type == evts.EVT_CLIENT_SUICIDE:
            p.die().act()
        elif type == evts.EVT_GAME_ROUND_START:
            if event.data.rounds == 1:
                self._mapstart = curtime
            self._roundstart = curtime
            # shuffle the teams before each new round (since doing so on map change doesn't seem to fly)
            if self.autoBalance and event.data.rounds % self.balanceIntervalRounds == 2: # first check will be on 2nd round
                self.report("EVT_GAME_ROUND_START(round %s): checkbalance" % event.data.rounds)
                self._nextcheck = min(curtime + 5, self._nextcheck)
            else:
                self.report("EVT_GAME_ROUND_START(round %s): do nothing" % event.data.rounds)
        elif type in [evts.EVT_GAME_EXIT, evts.EVT_GAME_MAP_CHANGE]:
            # we must not respond to this trigger twice for the same ending
            if curtime - self._lastreset < 30:
                return
            with self._control:
                for id, player in self._players.items():
                    if player.getTeam() == b3.TEAM_DROPPED:
                        del self._players[id]
            self._mapstart = 0
            self._roundstart = 0
            gType = self.console.game.gameType
            if self.autoShuffle and gType != "bm":
                self.report("EVT_GAME_EXIT: autoshuffle (gametype %s)" % gType)
                self.doShuffle(None, True)
            else:
                self.report("EVT_GAME_EXIT: checkbalance (gametype %s)" % gType)
                self._lastadjust = 0
                self._nextcheck = min(curtime + 6, self._nextcheck)
                self.checkBalance(True)
            # reset scores /after/ shuffling teams based on last map's scores
            # and clear out idle/departed players (which were retained until end of map to prevent gaming skill assessment)
            self._lastreset = curtime
            for id, player in self._players.items():
                player.reset()
            # we mustn't rebalance at the beginning of the next round
            self._lastadjust = curtime + min(30, self.balanceCooldownSeconds)
            self._nextcheck = curtime + min(50, self.balanceCooldownSeconds)
        elif type == evts.EVT_GAME_WARMUP:
            self._mapstart = 0
            self._roundstart = 0
            self.debug("warmup, disabling autobalance for %i seconds" % min(30, self.balanceCooldownSeconds))
            self._lastadjust = curtime + min(30, self.balanceCooldownSeconds)
            self._nextcheck = curtime + min(50, self.balanceCooldownSeconds)
            self.rebuildPlayerList()
        else:
            self.dumpEvent(event)
    
    # on event data, stupid B3 sends sometimes GUIDs, sometimes CIDs, who knows what else
    def findPlayer(self, data):
        self.verbose("findPlayer: search using '%s'" % data)
        
        try: # maybe we have GUID?
            return self._players[event.data]
        except:
            self.verbose("findPlayer: GUID lookup fail")
            pass
        
        try: # maybe we have CID?
            data = int(float(data))
            for k, p in self._players.items():
                self.verbose("findPlayer: %s==%s" % (k, p))
                self.verbose("findPlayer: comparing %s, %s on %s" % (p._cid, data, p._name))
                if p._cid == data:
                    return p
        except:
            self.verbose("findPlayer: CID lookup fail")
            pass
        return self._nullplayer
    
    # get the XlrPlayer correlating to a client
    def getPlayer(self, client=None):
        if not client:
            return self._nullplayer
        
        try:
            c = self._players[client.guid]
            if c.getName() != client.name:
                if c.isBot():
                    c.setName(client.name)
                    return c
                c = XlrPlayer(self.console, client, self.getWeakerTeam())
                with self._control:
                    self._players[client.guid] = c
                self.verbose("getPlayer name mismatch on guid (%s != %s)" % (c.getName(), client.name))
            type = "Bot" if c.isBot() else "Player"
            self.verbose("getPlayer returning %s %s[%s] (%s) for Client %s[%s] (%s)" % (type, c.getName(), c.getTeamName(), c.guid, client.name, c.getTeamName(client.team), client.guid))
        except:
            c = XlrPlayer(self.console, client, self.getWeakerTeam())
            self.verbose("Player %s (%s) didn't exist and had to be created" % (c.getName(), c.getId()))
            with self._control:
                self._players[client.guid] = c
        #c.checkTeam() # spammy, slows game and gets stuff expired on the queue
        return c
    
    def getWeakerTeam(self):
        with self._control:
            redcount = 0
            bluecount = 0
            redskill = 0
            blueskill = 0
            for k, p in self._players.items():
                if p.getTeam() == b3.TEAM_BLUE:
                    bluecount += 1
                    blueskill += p.getSkill()
                elif p.getTeam() == b3.TEAM_RED:
                    redcount += 1
                    redskill += p.getSkill()
            
            sPct = self.percentDiff(redskill, blueskill)
            cPct = self.percentDiff(redcount, bluecount)
            
            if sPct >= self.maxTeamRatio and abs(redcount - bluecount) >= self.maxTeamDiff:
                if redcount > bluecount:
                    return b3.TEAM_BLUE
                else:
                    return b3.TEAM_RED
            
            if redskill > blueskill:
                return b3.TEAM_BLUE
            else:
                return b3.TEAM_RED
    
    def updatePlayerList(self, init=False):
        self.verbose("Updating player list")
        #self.console.getPlayerTeams() #DAMMIT!  WE NEED THIS!
        with self._control:
            # set players with missing clients to TEAM_DROPPED
            for k, p in self._players.items():
                client = self.console.clients.getByGUID(k)
                if not client:
                    p.setTeam(b3.TEAM_DROPPED, False)
                elif init:
                    p.setTeam(client.team)
            
            # add unlisted players
            for c in self.console.clients.getList():
                self.getPlayer(c)
    
    def rebuildPlayerList(self):
        with self._control:
            for k, p in self._players.items():
                if p.getTeam() == b3.TEAM_DROPPED:
                    del self._players[p.guid]
            
            return self.updatePlayerList() # without thread safety issues, maybe this won't be necessary
            
            players = {}
            for c in self.console.clients.getList():
                try:
                    op = self._players[c.guid]
                except:
                    op = None
                if op and op.getName() == c.name:
                    np = op
                else:
                    np = XlrPlayer(self.console, c)
                # use op.getTeam() instead of c.team because b3 teams are fucked up (always "unknown")
                # means we have no catch for true team assignment sync drift
                tid = op.getTeam()
                if tid in [b3.TEAM_RED, b3.TEAM_BLUE] and tid != np.getTeam():
                    np.setTeam(tid, False)
                players[c.guid] = np
                try:
                    del self._players[c.guid]
                except:
                    pass
            for k, p in self._players.items():
                if p.getTeam() in [b3.TEAM_RED, b3.TEAM_BLUE, b3.TEAM_FREE]:
                    p.setTeam(b3.TEAM_DROPPED, False)
            for k, p in players.items():
                self._players[k] = p
    
    def countUnknownPlayers(self):
        n = 0
        for k, p in self._players.items():
            if p.getTeam() == b3.TEAM_UNKNOWN:
                n += 1
        return n
    
    def getCmd(self, cmd):
        cmd = 'cmd_%s' % cmd
        if hasattr(self, cmd):
            func = getattr(self, cmd)
            return func

    def report(self, msg):
        if self.debugMode:
            self.console.say(msg)
    
    def nameEvent(self, type):
        events = ['EVT_EXIT','EVT_STOP','EVT_UNKNOWN','EVT_CUSTOM','EVT_CLIENT_SAY','EVT_CLIENT_TEAM_SAY','EVT_CLIENT_PRIVATE_SAY','EVT_CLIENT_CONNECT','EVT_CLIENT_AUTH','EVT_CLIENT_DISCONNECT', 'EVT_CLIENT_UPDATE','EVT_CLIENT_KILL','EVT_CLIENT_GIB','EVT_CLIENT_GIB_TEAM','EVT_CLIENT_GIB_SELF','EVT_CLIENT_SUICIDE','EVT_CLIENT_KILL_TEAM','EVT_CLIENT_DAMAGE','EVT_CLIENT_DAMAGE_SELF','EVT_CLIENT_DAMAGE_TEAM','EVT_CLIENT_JOIN','EVT_CLIENT_NAME_CHANGE','EVT_CLIENT_TEAM_CHANGE','EVT_CLIENT_ITEM_PICKUP','EVT_CLIENT_ACTION','EVT_CLIENT_BAN','EVT_CLIENT_BAN_TEMP','EVT_CLIENT_KICK','EVT_GAME_ROUND_START','EVT_GAME_ROUND_END','EVT_GAME_WARMUP','EVT_GAME_EXIT']
        return events[type-1]
    
    def processMoveQueue(self):
        if len(self._players) == 0:
            return
        if not self.isTeamGame():
            return
        nRed = 0
        nBlue = 0
        with self._control:
            for k, p in self._players.items():
                if p.getTeam() == b3.TEAM_RED:
                    nRed += 1
                elif p.getTeam() == b3.TEAM_BLUE:
                    nBlue += 1
            for k, p in self._players.items():
                if p.getTeam() == b3.TEAM_UNKNOWN:
                    if nBlue > nRed:
                        p.setTeam(b3.TEAM_RED, False)
                        nRed += 1
                    else:
                        p.setTeam(b3.TEAM_BLUE, False)
                        nBlue += 1
                p.applyBalance()

    def intervalBalanceCheck(self):
        # self.console.verbose("intervalBalanceCheck...")
        curtime = time.time()
        if curtime - self._lastlistupdate > 60:
            self._lastlistupdate = curtime
            self.updatePlayerList()
        
        if self.autoBalance and self._nextcheck < curtime:
            gType = self.console.game.gameType
            self._nextcheck = curtime + self.balanceIntervalSeconds
            if gType in ["tdm", "ctf", "cah", None]:
                self.report("intervalBalanceCheck(gametype %s): checkbalance" % gType)
                try:
                    self.checkBalance()
                except Exception as err:
                    self.debug("intervalBalanceCheck err: %s" % err)
            else:
                self.report("intervalBalanceCheck(gametype %s): do nothing" % gType)
                self._checkresult = "no check on type %s" % gType
        else:
            self.debug("next balance check in %i seconds" % (self._nextcheck - curtime))
    
    def dumpEvent(self, event):
        self.debug('dumpEvent -- Type %s, Client %s, Target %s, Data %s',
            self.nameEvent(event.type), event.client, event.target, event.data)
    
    def reportBalance(self, client):
        with self._control:
            ncount = 0
            dcount = 0
            bcount = 0
            scount = 0
            hcount = 0
            redcount = 0
            bluecount = 0
            redskill = 0
            blueskill = 0
            for k, p in self._players.items():
                if p.getTeam() == b3.TEAM_BLUE:
                    bluecount += 1
                    blueskill += p.getSkill()
                elif p.getTeam() == b3.TEAM_RED:
                    redcount += 1
                    redskill += p.getSkill()
                elif p.getTeam() == b3.TEAM_SPEC:
                    scount += 1
                elif p.getTeam() == b3.TEAM_UNKNOWN:
                    ncount += 1
                elif p.getTeam() == b3.TEAM_DROPPED:
                    dcount += 1
                if p.getTeam() not in [b3.TEAM_RED, b3.TEAM_BLUE]:
                    continue
                if p.isBot():
                    bcount += 1
                else:
                    hcount += 1
            
            sPct = self.percentDiff(redskill, blueskill)
            tPct = self.percentDiff(redcount, bluecount)
            tDiff = abs(redcount - bluecount)
            lastcheck = time.time() - self._lastcheck
            nextcheck = self._nextcheck - time.time()
            lastadjust = time.time() - self._lastadjust
            client.message("^5Balance Report")
            client.message("  players: ^5%ihum+%ibot, %iplay+%ispec+%ileft, %i?" % (hcount, bcount, redcount+bluecount, scount, dcount, ncount))
            client.message("  skill: ^5red=%i, blue=%i (diff %ipct)" % (redskill, blueskill, sPct*100))
            client.message("  count: ^5red=%i, blue=%i (diff %i, %ipct)" % (redcount, bluecount, tDiff, tPct*100))
            client.message("  last balance^5: %i seconds ago (%s)" % (lastadjust, self._adjustresult))
            client.message("  last check^5: %i seconds ago (%s)" % (lastcheck, self._checkresult))
            client.message("  next check^5: %i seconds" % nextcheck)
    
    # check if current or proposed team arrangement is fair (with higher standard for manual checks, identified by supplying the client that triggered them)
    def isBalanced(self, redlist, bluelist, redskill=None, blueskill=None, client=False):
        # use the potentially higher "auto-balance" thresholds if this isn't a manual check
        if client:
            mTeamDiff = self.maxTeamDiff
            mTeamRatio = self.maxTeamRatio
            mSkillRatio = self.maxSkillRatio
        else:
            mTeamDiff = self.maxAutoTeamDiff
            mTeamRatio = self.maxAutoTeamRatio
            mSkillRatio = self.maxAutoSkillRatio
    
        # we have to balance so players in unknown team will be assigned a team (so we can catch it on EVT_CLIENT_TEAM_CHANGE)
        for k, p in self._players.items():
            if p.getTeam() == b3.TEAM_UNKNOWN:
                return False
        
        # Check for imbalances by player counts
        nRed = len(redlist)
        nBlue = len(bluelist)
        
        tDiff = abs(nRed - nBlue)
        tPct = self.percentDiff(nRed, nBlue)
        if client and self.debugBalance:
            client.message("Team counts: %i, %i (%i, %i percent difference)" % (nRed, nBlue, tDiff, tPct*100))
        self.verbose("team delta is %i (%f)" % (tDiff, tPct))
        # player count balancing (must violate both to be considered imbalanced
        if tDiff > mTeamDiff and tPct > mTeamRatio:
            return False
        
        # Check for imbalances by per-team player skill sums
        if not redskill or not blueskill:
            redskill = 0
            blueskill = 0
            for p in redlist:
                redskill += p.getSkill()
            for p in bluelist:
                blueskill += p.getSkill()
        # skill sums should not differ by more than 25% (10% on manual check)
        sPct = self.percentDiff(redskill, blueskill)
        if client and self.debugBalance:
            client.message("Team skills: %i, %i (%i percent difference)" % (redskill, blueskill, sPct*100))
        self.verbose("team skill difference is %f" % sPct)
        if sPct > mSkillRatio:
            return False
        return True
    
    def percentDiff(self, n1, n2):
        if n1 == 0 and n2 == 0:
            return 0
        if n1 == 0 or n2 == 0:
            return 100
        diff = float(abs(n1 - n2))
        return max(diff / n1, diff / n2)
    
    # check if balance is markedly poor (with higher standards for manual checks) and if necessary, fix
    def checkBalance(self, client=None):
        self.debug("checking balance")
        with self._control:
            for k, p in self._players.items():
                p.checkTeam()
            curtime = time.time()
            self._lastcheck = curtime
            msg = False
            mustWait = False
            checkWaitSeconds = self.balanceCooldownSeconds - (curtime - self._lastadjust)
            
            # self.console.game stores times in minutes (we use seconds) and doesn't even have them after a restart
            try:
                timeLimit = float(self.console.game.timeLimit) * 60
            except:
                timeLimit = 1200 # just guess that it's 20 minutes, long enough that the 2/3rds rule won't take effect
            if self._mapstart > 0:
                timeSpent = curtime - self._mapstart
                timeLeft = timeLimit - timeSpent
            else:
                timeSpent = 0
                timeLeft = 0
            endTime = min(2.0 * self.balanceDisableSecondsEOM, timeLimit / 3.0)
            noCheckStartTime = max(min(self.balanceDisableSecondsBOM, 120), 30)
            
            if str(client) == "True":
                lvl = 100
            elif client != None:
                #client.message("timeLimit=%f, timeLeft=%f, endTime=%f" % (timeLimit, timeLeft, endTime))
                lvl = client._maxLevel
            else:
                lvl = -1
            
            # Idle map check
            nPlayers = 0
            for k, p in self._players.items():
                nPlayers += 1
            if nPlayers == 0 or (nPlayers < 2 and self.isTeamGame()):
                msg = "Nobody seems to be here"
                if not self._emptysince > 0:
                    self._emptysince = time.time()
                if lvl > -1 and str(client) != "True":
                    client.message(msg)
                self._checkresult = msg
                self._nextcheck = time.time() + 10
                self.debug(msg)
                self.report(msg)
                if self._emptysince and time.time() - self._emptysince > self.maxIdleMap:
                    self.console.write("cyclemap")
                return False
            self._emptysince = 0;
            
            # check that balance conditions are met
            if not self.autoBalance and lvl == -1:
                msg = "Automatic rebalancing disabled"
            elif timeSpent > 0 and timeSpent < noCheckStartTime:
                msg = "Cannot rebalance a game near the beginning of a round"
                self._nextcheck = self._mapstart + max(self.balanceCooldownSeconds, 90)
                #mustWait = True # too disruptive here
            # all checks disabled during the last 2*cooldown seconds of a round or last 1/3rd of the round (whichever is smaller)
            elif timeSpent > 0 and timeLeft < endTime and lvl < 100:
                msg = "Cannot rebalance a game near the end of a round"
                #mustWait = True # too disruptive here
            # manual checks are allowed at halfway through cooldown
            elif checkWaitSeconds > self.balanceCooldownSeconds/2.0 and lvl >= 60:
                msg = "Teams were just rebalanced--wait %i seconds" % (checkWaitSeconds+2 - self.balanceCooldownSeconds/2.0)
                self._nextcheck = min(curtime + checkWaitSeconds + 1, self._nextcheck)
                mustWait = True
            elif checkWaitSeconds > 0 and lvl < 60:
                msg = "Teams were just rebalanced--wait %i seconds" % (checkWaitSeconds+1)
                self._nextcheck = min(curtime + checkWaitSeconds + 1, self._nextcheck)
                mustWait = True
            elif not self.isTeamGame():
                msg = "Cannot rebalance a non-team game"
            
            
            redskill = 0
            blueskill = 0
            if mustWait or not msg:
                redteam = []
                blueteam = []
                # first, remove inactive players from the roster so they don't pollute the team assessment
                for k, p in self._players.items():
                    p.checkActive()
                # then build lists according to current teams (getTeam() returns queued changes over current values)
                for k, p in self._players.items():
                    if p.getTeam() == b3.TEAM_BLUE:
                        blueteam.append(p)
                        blueskill += p.getSkill()
                    elif p.getTeam() in [b3.TEAM_RED, b3.TEAM_UNKNOWN]: # capturing unknowns here
                        redteam.append(p)
                        redskill += p.getSkill()
                
                # can't autobalance with so few players, nor do we need to
                if len(redteam) + len(blueteam) < self.minPlayers:
                    msg = "Not enough players to balance (need %i)" % self.minPlayers
                elif self.isBalanced(redteam, blueteam, redskill, blueskill, client):
                    msg = "Teams are already balanced"
            
            pctDiff = self.percentDiff(redskill, blueskill)
            if mustWait and pctDiff > self.immediateSkillRatio:
                self.console.say("Team skills imbalanced by %i percent, forcing immediate rebalance" % (pctDiff*100))
                msg = False
                client = True
                
            if msg:
                if lvl > -1 and str(client) != "True":
                    client.message(msg)
                self._checkresult = msg
                self.debug(msg)
                self.report(msg)
                return False
            self._checkresult = "rebalance required"
            # if lvl > -1:
                # client.message(self._checkresult)
            
            self._nextcheck = curtime + self.balanceIntervalSeconds
            success = self.findBalance(redteam, blueteam, client)
            if success:
                movingplayers = False
                for p in redteam:
                    if p.getTeam() != b3.TEAM_RED and not p.isBot():
                        movingplayers = True
                        break
                for p in blueteam:
                    if p.getTeam() != b3.TEAM_BLUE and not p.isBot():
                        movingplayers = True
                        break
                if movingplayers:
                    self.console.saybig("Adjusting team balance")
                self.applyTeams(redteam, blueteam)
                self._checkresult = "rebalance applied"
            else:
                self._checkresult = "rebalance failed"
            # if lvl > -1:
                # client.message(self._checkresult)
            return True
    
    def isTeamGame(self):
        try:
            x = len(self.console.getCvar('g_redteamlist').getString())
            x = len(self.console.getCvar('g_blueteamlist').getString())
            return True
        except:
            return False
    
    # generate loosely skill-based balanced player distribution
    def doShuffle(self, client=None, endgame=False):
        if not self.isTeamGame():
            if client != None:
                client.message("Cannot shuffle, not a team game")
            return False
        
        with self._control:
            if endgame:
                self.console.say("Shuffling Teams")
            else:
                self.console.saybig("Shuffling Teams") # this isn't seen during endgame because the score screen is up
            self.debug("shuffling teams")
            teams = {b3.TEAM_RED: [], b3.TEAM_BLUE: []}
            tids = [b3.TEAM_RED, b3.TEAM_BLUE] if random.randint(0, 1) else [b3.TEAM_BLUE, b3.TEAM_RED]
            players = []
            for id, p in self._players.items():
                t = p.getTeam()
                if t in [b3.TEAM_FREE, b3.TEAM_RED, b3.TEAM_BLUE, b3.TEAM_UNKNOWN]:
                    players.append(p)
                    p.allowRebalance()
            str = ""
            for p in players:
                str += "%s {%s}, " % (p.getName(), p.getSkill())
            self.debug("players: %s" % str)
            if len(players) == 0:
                self.debug("no players to shuffle")
                return False
            players.sort() # sorted by skill, highest first
            
            str = ""
            for p in players:
                str += "%s {%s}, " % (p.getName(), p.getSkill())
            self.debug("players-sorted: %s" % str)
            
            #players.sort(key=lambda p: p.getSkill())
            # p1 in t1, p2 and p3 in t2, p4 and p5 in t1...
            teams[tids[0]].append(players[0])
            i = 1
            t = 1
            while i < len(players):
                teams[tids[t]].append(players[i])
                i += 1
                if i < len(players):
                    teams[tids[t]].append(players[i])
                    i += 1
                    t = (t + 1) % 2
            
            self.findBalance(teams[tids[0]], teams[tids[1]], True)
            self.applyTeams(teams[tids[0]], teams[tids[1]])
    
    # fine-tune current or proposed team assignments (edit supplied lists and return boolean success)
    def findBalance(self, redlist, bluelist, force=False):
        str = ""
        curtime = time.time()
        for p in redlist:
            str += "%s{%s}, " % (p.getName(), p.getSkill())
        str2 = ""
        for p in bluelist:
            str2 += "%s{%s}, " % (p.getName(), p.getSkill())
        self.debug("distributed teams: *T1*: %s *T2* %s" % (str, str2))
        
        # start at 1 to avoid division by 0
        sRed = 0
        sBlue = 0
        for p in redlist:
            sRed += p.getSkill()
        sRedOld = sRed
        for p in bluelist:
            sBlue += p.getSkill()
        sBlueOld = sBlue
        nRed = len(redlist)
        nBlue = len(bluelist)
        nDiff = self.percentDiff(nRed, nBlue)
        sDiff = self.percentDiff(sRed, sBlue)
        
        oldresult = self._adjustresult
        self._adjustresult = "ok"
        
        # first, balance player count using the players who've gone the longest without an auto-team switch (never switched = forever = first picked)
        s = 0
        mtd = self.maxTeamDiff
        mtr = self.maxTeamRatio
        while nRed - nBlue > mtd or (sRed - sBlue > s and nRed >= nBlue):
            if s > 0:
                redlist.remove(p)
                nRed -= 1
                sRed -= s
                bluelist.append(p)
                nBlue += 1
                sBlue += s
            p = self.findOldest(redlist, b3.TEAM_BLUE)
            s = p.getSkill()
        s = 0
        while nBlue - nRed > mtd or (sBlue - sRed > s and nBlue >= nRed):
            if s > 0:
                bluelist.remove(p)
                nBlue -= 1
                sBlue -= s
                redlist.append(p)
                nRed += 1
                sRed += s
            p = self.findOldest(bluelist, b3.TEAM_RED)
            s = p.getSkill()
        
        # then balance the skill (likely) without altering player counts, using swaps of nearest skill match
        i = 0
        while self.percentDiff(sRed, sBlue) > mtr:
            # push over the player with closest skill level to the difference.
            if sRed > sBlue:
                p = self.findClosest(redlist, sRed - sBlue, b3.TEAM_BLUE)
                s = p.getSkill()
                redlist.remove(p)
                sRed -= s
                nRed -= 1
                bluelist.append(p)
                sBlue += s
                nBlue += 1
                
                if sBlue - sRed > self._nullplayer.minSkill or (nBlue - nRed > mtd and self.percentDiff(nRed, nBlue) > mtr):
                    p = self.findLowest(bluelist)
                    s = p.getSkill()
                    bluelist.remove(p)
                    sBlue -= s
                    nBlue -= 1
                    redlist.append(p)
                    sRed += s
                    nRed += 1
            else:
                p = self.findClosest(bluelist, sBlue - sRed, b3.TEAM_RED)
                s = p.getSkill()
                bluelist.remove(p)
                sBlue -= s
                nBlue -= 1
                redlist.append(p)
                sRed += s
                nRed += 1
                
                if sRed - sBlue > self._nullplayer.minSkill or (nRed - nBlue > mtd and self.percentDiff(nRed, nBlue) > mtr):
                    p = self.findLowest(redlist)
                    s = p.getSkill()
                    redlist.remove(p)
                    sRed -= s
                    nRed -= 1
                    bluelist.append(p)
                    sBlue += s
                    nBlue += 1
            
            i += 1
            if i > 80:
                self.debug("Failed to find balance after %i iterations" % i)
                #self.console.say("Failed to find balance after %i iterations" % i)
                self._adjustresult = "too many steps"
                if self.percentDiff(sRed, sBlue) > (self.maxAutoSkillRatio + self.maxSkillRatio) / 2.0 and not force:
                    self._toolongcount += 1
                    if self._toolongcount < 3:
                        self.console.say("Team adjustment failed (too many iterations)")
                        self._nextcheck = curtime + 10
                        return False
                break
        
        
        moves = self.countMoves(redlist, bluelist)
        humanMoves = self.countMoves(redlist, bluelist, True)
        unknowns = self.countUnknownPlayers()
        totalPlayers = len(redlist) + len(bluelist)
        
        # do not apply team switches during a rebalance not being applied to a shuffle
        if unknowns + moves == 0 and not force:
            self.console.say("Team adjustment failed (couldn't find better teams)")
            self._lastadjust = curtime - math.floor(self.balanceCooldownSeconds / 2.0) # prevent repeat attempt for half of min-interval
            self._adjustresult = "no moves"
            self._nextcheck = curtime + 30
            return False
        
        if unknowns == 0 and humanMoves >= totalPlayers/2.8 and not force and totalPlayers >= 6 and nDiff < self.maxTeamRatio:
            self._toomanycount += 1
            if self._toomanycount < 3:
                self._lastadjust = curtime - math.floor(self.balanceCooldownSeconds / 3.0) # prevent repeat attempt for half of min-interval
                self._adjustresult = "too many moves"
                self._nextcheck = curtime + 20
                self.console.say("Team adjustment failed (moving too many players)")
                self.debug("Balance failed...tried to switch too many players")
                return False
        
        # do not apply team switches if the skill difference is actually worse than we had before
        new_sDiff = self.percentDiff(sRed, sBlue)
        if new_sDiff >= sDiff:
            self._adjustresult = "non-improvement"
            if oldresult != "non-improvement" and not force:
                self.console.say("Team adjustment failed (non-improved result)") #, %i:%i -> %i:%i)" % (sRedOld, sBlueOld, sRed, sBlue))
            self._nextcheck = curtime + 15
            return False
        self._toolongcount = 0
        self._toomanycount = 0
        
        # swap teams if switches exceed 50% of players (note this bypasses bounce protection, and also doesn't work because the lists are object references)
        # if moves > totalPlayers / 2:
            # swap = redlist
            # redlist = bluelist
            # bluelist = swap
        
        str = ""
        for p in redlist:
            str += "%s{%s}, " % (p.getName(), p.getSkill())
        str2 = ""
        for p in bluelist:
            str2 += "%s{%s}, " % (p.getName(), p.getSkill())
        self.debug("balanced teams: *T1*: %s *T2* %s" % (str, str2))
        if i <= 80:
            self.debug("Balance found in %i iterations" % i)
        return True
    
    def prepMapChange(self):
        '''make sure we're using the right maplist and botsupport settings for the next map'''
    
    def cycleMap(self):
        '''change the map'''
        self.prepMapChange()
    
    def countMoves(self, redlist, bluelist, humansOnly=False):
        moves = 0
        for p in redlist:
            if p.getTeam() != b3.TEAM_RED:
                if not humansOnly or not p.isBot():
                    moves += 1
        for p in bluelist:
            if p.getTeam() != b3.TEAM_BLUE:
                if not humansOnly or not p.isBot():
                    moves += 1
        return moves
    
    # find player in the list that best matches the target skill level
    def findClosest(self, playerList, targetSkill, forteam=b3.TEAM_SPEC):
        # find nearest match in excess of targetSkill
        hPlayer = playerList[0]
        hDiff = hPlayer.getSkill() - targetSkill
        for p in playerList:
            if self.balancePreventBouncing and p.getTeam() != forteam and not p.canRebalance():
                continue
            diff = p.getSkill() - targetSkill
            if diff > 0 and diff < hDiff:
                hPlayer = p
                hDiff = diff
        
        #find nearest match short of targetSkill
        lPlayer = playerList[0]
        lDiff = targetSkill - lPlayer.getSkill()
        for p in playerList:
            if self.balancePreventBouncing and p.getTeam() != forteam and not p.canRebalance():
                continue
            diff = targetSkill - p.getSkill()
            if diff > 0 and diff < lDiff:
                lPlayer = p
                lDiff = diff
        
        #if both matches are similar (within 10%), use the "over" (otherwise, purely nearest match)
        dDiff = self.percentDiff(hDiff, lDiff)
        return hPlayer if dDiff < .1 or hDiff < lDiff else lPlayer
    
    def findOldest(self, playerList, forteam=b3.TEAM_SPEC):
        match = playerList[0]
        for p in playerList:
            if self.balancePreventBouncing and p.getTeam() != forteam and not p.canRebalance():
                continue
            if p.getLastAdjust() < match.getLastAdjust():
                match = p
        return match
    
    # find the worst (lowest skilled) player on the list
    def findLowest(self, playerList):
        match = playerList[0]
        for p in playerList:
            if p.getSkill() < match.getSkill():
                match = p
        return match
    
    # move players into the proposed team assignments
    def applyTeams(self, redlist, bluelist):
        self._lastadjust = time.time()
        self._adjustresult += ", applied"
        self.debug("APPLYING TEAM SWITCHES")
        moves = [0, 0, 0] # null moves, bot moves, human moves
        finished = False
        while not finished:
            count = moves[1] + moves[2]
            moves[self.applySingle(redlist, b3.TEAM_RED)] += 1
            moves[self.applySingle(bluelist, b3.TEAM_BLUE)] += 1
            if count == moves[1] + moves[2]:
                finished = True
        
        if moves[1] + moves[2] > 0:
            msg = "teamswitching %i of %i players" % (moves[1] + moves[2], len(redlist) + len(bluelist))
            if moves[1]:
                msg += " (%i bots)" % (moves[1])
            self.console.say(msg)
        
        self._lastadjust = time.time() # repeat because the process can sometimes be slow, especially when held up by reporting output
        self.debug("REBALANCING DONE")
    
    # move a player on list to team (that isn't already) return false if none
    def applySingle(self, list, team):
        for p in list:
            if p.setTeam(team, False): # manual = false
                return 1 if p.isBot() else 2
        return 0
    
    # shuffle the player teams
    def cmd_shuffle(self, data, client, cmd=None):
        self.getPlayer(client).setHuman()
        self.doShuffle(client)
    
    # ensure teams are balanced (with higher standard than auto-checks)
    def cmd_balance(self, data, client, cmd=None):
        self.getPlayer(client).setHuman()
        self.checkBalance(client)
    
    def cmd_balancereport(self, data, client, cmd=None):
        self.getPlayer(client).setHuman()
        self.reportBalance(client)
    
    def cmd_scores(self, data, client, cmd=None):
        client.message("^5Scores:")
        for k, p in self._players.items():
            if p._team == b3.TEAM_DROPPED:
                continue
            color = "1" if p.getTeam() == b3.TEAM_RED else "4" if p.getTeam() == b3.TEAM_BLUE else "7"
            client.message("  ^%s%s^7 (@%i): ^5%i ^7(%i, %i/%i)" % (color, p.getName(), p._xlrid, p.getSkill(), p._xlrskill, p._kills, p._deaths))
    
    def cmd_listplayers(self, data, client, cmd=None):
        self.getPlayer(client).setHuman()
        rbnames = []
        rhnames = []
        bbnames = []
        bhnames = []
        sbnames = []
        shnames = []
        dnames = []
        for k, p in self._players.items():
            n = p.getName()
            t = p.getTeam()
            b = p.isBot()
            if t == b3.TEAM_RED:
                if b:
                    rbnames.append(n)
                else:
                    rhnames.append(n)
            elif t == b3.TEAM_BLUE:
                if b:
                    bbnames.append(n)
                else:
                    bhnames.append(n)
            elif t == b3.TEAM_SPEC:
                if b:
                    sbnames.append(n)
                else:
                    shnames.append(n)
            elif t == b3.TEAM_DROPPED:
                dnames.append(n)
        client.message("%i Red Bots: ^5%s" % (len(rbnames), ', '.join(rbnames)))
        client.message("%i Red Humans: ^5%s" % (len(rhnames), ', '.join(rhnames)))
        client.message("%i Blue Bots: ^5%s" % (len(bbnames), ', '.join(bbnames)))
        client.message("%i Blue Humans: ^5%s" % (len(bhnames), ', '.join(bhnames)))
        client.message("%i Spec Bots: ^5%s" % (len(sbnames), ', '.join(sbnames)))
        client.message("%i Spec Humans: ^5%s" % (len(shnames), ', '.join(shnames)))
        client.message("%i Dropped: ^5%s" % (len(dnames), ', '.join(dnames)))
    
    # toggle automatically shuffling teams before each map
    def cmd_autoshuffle(self, data, client, cmd=None):
        if not client:
            return
        self.getPlayer(client).setHuman()
        report = True
        
        if data:
            param = data.lower()
            if param not in ('off', 'on'):
                client.message("Invalid parameter. Expecting 'on' or 'off'")
                report = False
            elif param == 'off':
                self.autoShuffle = False
            elif param == 'on':
                self.autoShuffle = True
        if report or not data:
            state = 'enabled' if self.autoShuffle else 'disabled'
            client.message('XLRTeams: autoshuffle is currently %s.' % state)
    
    # toggle automatically maintaining team balance
    def cmd_autobalance(self, data, client, cmd=None):
        if not client:
            return
        self.getPlayer(client).setHuman()
        report = True
        
        if data:
            param = data.lower()
            if param not in ('off', 'on'):
                client.message("Invalid parameter. Expecting 'on' or 'off'")
                report = False
            elif param == 'off':
                self.autoBalance = False
            elif param == 'on':
                self.autoBalance = True
        if report or not data:
            state = 'enabled' if self.autoBalance else 'disabled'
            client.message('XLRTeams: autobalance is currently %s.' % state)
    
    # toggle preventing players from getting bounced between teams
    def cmd_preventbouncing(self, data, client, cmd=None):
        if not client:
            return
        self.getPlayer(client).setHuman()
        report = True
        
        if data:
            param = data.lower()
            if param not in ('off', 'on'):
                client.message("Invalid parameter. Expecting 'on' or 'off'")
                report = False
            elif param == 'off':
                self.balancePreventBouncing = False
            elif param == 'on':
                self.balancePreventBouncing = True
        if report or not data:
            state = 'enabled' if self.balancePreventBouncing else 'disabled'
            client.message('XLRTeams: preventbouncing is currently %s.' % state)
    
# /XlrteamsPlugin    